/*************************************************************************
Title:    USB interface for HD44780 displays with Crystal Fontz emulation
Author:   Pontus Frssander <pontusf@kth.se> http://www.e.kth.se/~pontusf
Software: AVR-GCC 3.3
Hardware: any AVR with built-in UART and two PWM channels, 
	  tested on ATtiny2313 with 8 MHz external ceramic resonator.

DESCRIPTION:
          This program emulates a Crystal Fontz serial display over USB.
          The emulation allows the user to use pre-existing programs to
          talk to the LCD. There is support for setting the contrast level
          and LED backlight intensity in software.

*************************************************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/pgmspace.h>
#include <stdbool.h>
#include "spin_delay.h"
#include "uart.h"
#include "lcd.h"


#define XTAL_CPU                 8000000     /* 8 MHz */
#define MCU_CLK                  8000000
#define UART_BAUD_RATE           9600        /* 9600 baud */
#define LCD_NUMROWS              4           /* 2 rows */
#define LCD_NUMCOLS              20          /* 16 columns */

#define CMD_NOTHING              0
#define CMD_CURSORCOLUMN         1
#define CMD_CURSORROW            2
#define CMD_SETBACKLIGHT		 3
#define CMD_SETCONTRAST			 4

#define OC_contrast PB2
#define OC_led PB3
#define DDROC DDRB
#define OCR_contrast OCR0A
#define OCR_led OCR1A


volatile uint16_t pwm_led; 
volatile uint8_t pwm_contrast; 

/* Custom characters. (8 * 8 = 64 bytes). */

static const PROGMEM unsigned char customchars[64] = {
	
	/*  [0] Stort  */
	/*  000  __#__  */  004,
	/*  001  _____  */  000,
	/*  002  _###_  */  016,
	/*  003  #___#  */  021,
	/*  004  #####  */  037,
	/*  005  #___#  */  021,
	/*  006  #___#  */  021,
	/*  007  _____  */  000,
	
	/*  [1] Stort  */
	/*  010  _#_#_  */  012,
	/*  011  _____  */  000,
	/*  012  _###_  */  016,
	/*  013  #___#  */  021,
	/*  014  #####  */  037,
	/*  015  #___#  */  021,
	/*  016  #___#  */  021,
	/*  017  _____  */  000,
	
	/*  [2] Litet  */
	/*  020  __#__  */  004,
	/*  021  _____  */  000,
	/*  022  _###_  */  016,
	/*  023  ____#  */  001,
	/*  024  _####  */  017,
	/*  025  #___#  */  021,
	/*  026  _####  */  017,
	/*  027  _____  */  000,

	/*  [3] Litet  */
	/*  030  _____  */  000,
	/*  031  _#_#_  */  012,
	/*  032  _____  */  000,
	/*  033  _###_  */  016,
	/*  034  #___#  */  021,
	/*  035  #___#  */  021,
	/*  036  _###_  */  016,
	/*  037  _____  */  000,

	/*  [4] Backsl. */
	/*  040  _____  */  000,
	/*  041  #____  */  020,
	/*  042  _#___  */  010,
	/*  043  __#__  */  004,
	/*  044  ___#_  */  002,
	/*  045  ____#  */  001,
	/*  046  _____  */  000,
	/*  047  _____  */  000,


	/*  [5] Playsym */
	/*  060  _#___  */  010,
	/*  061  _##__  */  014,
	/*  062  _###_  */  016,
	/*  063  _####  */  017,
	/*  064  _###_  */  016,
	/*  065  _##__  */  014,
	/*  066  _#___  */  010,
	/*  067  _____  */  000,

	/*  [6]  Pause  */
	/*  070  _____  */  000,
	/*  071  _#_#_  */  012,
	/*  072  _#_#_  */  012,
	/*  073  _#_#_  */  012,
	/*  074  _#_#_  */  012,
	/*  075  _#_#_  */  012,
	/*  076  _____  */  000,
	/*  077  _____  */  000,
	
	/*  [7] Eurosym */
	/*  050  __###  */  007,
	/*  051  _#___  */  010,
	/*  052  ####_  */  036,
	/*  053  _#___  */  010,
	/*  054  ####_  */  036,
	/*  055  _#___  */  010,
	/*  056  __###  */  007,
	/*  057  _____  */  000
};

/* PWM interrupt routines */
SIGNAL (SIG_TIMER1_OVF) 
{
    OCR_led = pwm_led;				/* Set backlight level */
}

SIGNAL (SIG_TIMER0_OVF) 
{
    OCR_contrast = pwm_contrast;	/* Set contrast level */
}

void ioinit (void) /* Initialize PWM timers*/
{

	/* Initialize backlight PWM */
	
	 /* led timer (timer1) is 10-bit PWM */
    TCCR1A = _BV (WGM10) | _BV (WGM11) | _BV (COM1A1);
	
	/* led timer (timer1) ckdiv = 8 */
    TCCR1B = _BV (CS11);
	
	/* set PWM value to 0 */
    OCR_led = 0;
	
	/* enable PB3 as led output */
    DDRB |= _BV (OC_led);
	
	/* backlight PWM initialized */
	
	/* initialize contrast PWM */
	
	/* Contrast timer (timer 0) is 8-bit pwm */
	TCCR0A = _BV (WGM00) | _BV (WGM01) | _BV (COM0A1);
	
	/* Contrast timer (timer 0)  ckdiv = 64 */
    TCCR0B = (_BV (CS01) | _BV (CS00));
	
	/* set PWM value to 0 */
    OCR_contrast = 0;
	
	/* enable PB2 as contrast output */
    DDRB |= _BV (OC_contrast);
	
	/* Enable timer0 and timer1 interrupts */
	timer_enable_int (_BV (TOIE1) | _BV (TOIE0));
	
	/* now enable interrupts */
	sei();
}

void lcdstart (void){
	/* Initialize display, cursor off */
	lcd_init(LCD_DISP_ON);
	
	/* Load custom characters into LCD */
	lcd_command(_BV(LCD_CGRAM));	/* set CG RAM start address 0 */
	for(int i=0; i<64; i++)		/* Loop through the bitmap array */
       {
			lcd_data(pgm_read_byte_near(&customchars[i]));
       }
	   
	/* Home cursor */
	lcd_home();
	
	/* clear display */
	lcd_clrscr(); 		
	
	/* Boot screen */
    lcd_puts("Starting up");
	pwm_contrast = 10;
	pwm_led = 128;
}

int main(void)
{
	unsigned char cursorcolumn = 0;					/* Used when moving cursor to new column (and then row) */
	unsigned int c;									/* Received data from UART 								*/
	unsigned char skipbytes = 0;						/* Number of bytes being skipped. (Max 255)				*/
	unsigned char curcmd = CMD_NOTHING;				/* Current command being evaluated. 					*/
	const char *hexdigits = PSTR("0123456789ABCDEF");	/* Custom error code handling 							*/
	ms_spin(500);
	DDRB = 0xFF;
	PORTB = 0x00;
	
	ioinit (); /* Initialize PWM for contrast and backlight */
 
	lcdstart();	 /* Fire up the LCD */
	
	/*
	 *  Initialize UART library, calculate the argument
	 *  yourself from the datasheet.
	 */
	uart_init(51); 
	
	while(1)
	{
    	
		/*
		 * Get received character from ringbuffer
		 * uart_getc() returns in the lower byte the received character and 
		 * in the higher byte (bitmask) the last receive error
		 * UART_NO_DATA is returned when no data is available.
		 */
		while ( (c = uart_getc()) == UART_NO_DATA )
			;

		/* Are we skipping bytes? If so, do it. */
		if (skipbytes)
		{
			skipbytes--;
			continue;
		}

		/*
		 * new data available from UART
		 * if there was a receive error, it is encoded in the 8 most significant bits.
		 */
		if ( c & 0xff00 )
		{
			lcd_puts_P("UART ERROR ");
			lcd_putc(PRG_RDB(&hexdigits[c >> 12]));
			lcd_putc(PRG_RDB(&hexdigits[(c >> 8) & 0x0f]));
			lcd_putc('\n');
			c &= 0xff; /* Strip of the error bits, and continue handling. */
		}

		if (curcmd != CMD_NOTHING) switch (curcmd)
		{
		case CMD_CURSORCOLUMN: 				/* Used when pointing the cursor to a specific  */
			cursorcolumn = c;				/* column and row                               */
			curcmd = CMD_CURSORROW;
			break;

		case CMD_CURSORROW:					/* Used when pointing the cursor to a specific  */
			lcd_gotoxy(cursorcolumn,c);		/* column and row                               */	
			curcmd = 0;
			break;
		case CMD_SETBACKLIGHT:				/* Set backlight level 							*/
			pwm_led = c*10;					/* received value is 0-100 (100=brightest) but  */
			curcmd = 0;						/* valid PWM values are 0 to 1024 				*/
			break;
		case CMD_SETCONTRAST:				/* Set contrast level 							*/
			pwm_contrast = (50 - (c/2));	/* received value is 0-100 (100=darkest) but 	*/
			curcmd = 0;						/* valid PWM values are 0 to 255 				*/
			break;
		default: // Danger! Will Robinson! curcmd set to a spooky value
			lcd_puts_P("CURCMD ERROR ");
		//	lcd_putc(PRG_RDB(&hexdigits[curcmd >> 4]));		//This took up too much space.
		//	lcd_putc(PRG_RDB(&hexdigits[curcmd & 0x0f]));	//This took up too much space.
		//	lcd_putc('\n');									//This took up too much space.
			break;
		}

		/*
		 * NOTE NOTE NOTE: Do not remove any of the 0..31 command checks! 
		 * The 'default' entry at the bottom of the switch
		 * relies on knowing that the character is not a control character.
		 */
		else switch (c) {
		case 0: // Null command
			break;

		case 1: // Cursor home
			lcd_home();
			break;

		case 2: // Shut off display but don't clear it
			lcd_command(LCD_DISP_OFF);
			break;

		case 3: // Restore display
		case 4: // Hide cursor
			lcd_command(LCD_DISP_ON); 
			break;

		case 5: // Show underline cursor
			lcd_command(LCD_DISP_ON_CURSOR);
			break;

		case 6: // Show block cursor
		case 7: // Show inverting cursor. Inverted not supported on HD44780 displays!!
			lcd_command(LCD_DISP_ON_CURSOR_BLINK);
			break;

		case 8: // Backspace
			lcd_command(LCD_MOVE_CURSOR_LEFT);
			lcd_putc(' ');
			lcd_command(LCD_MOVE_CURSOR_LEFT);

			break;

		case 9: // Control the boot screen. Not implemented.
			skipbytes = 1;
			break;

		case 10: // Line feed
			lcd_puts("\n");
			break;

		case 11: // Delete in place
			lcd_putc(' ');
			lcd_command(LCD_MOVE_CURSOR_LEFT);
			break;

		case 12: // Form feed (clear display)
			lcd_clrscr();
			break;

		case 13: // Carriage return FIXME: Implement this somehow.
			//lcd_putc('\r');
			break;

		case 14: // Backlight control
			curcmd = CMD_SETBACKLIGHT; //set the backlight level next iteration
			break;
			
		case 15: // Contrast control
			curcmd = CMD_SETCONTRAST;  //set contrast level next iteration
			break;
			
		case 16: // Not used
			break;
			
		case 17: // Set cursor position (column and row)
			curcmd = CMD_CURSORCOLUMN;	//set column next iteration (and after that, set row)
			break;

		case 18: // Horizontal bar graph. Not implemented.
			skipbytes = 6;
			break;

		case 19: // Scroll ON. Not implemented. FIXME: implement this
			break;

		case 20: // Scroll OFF. Not implemented. FIXME: implement this
			break;

		case 21: // Set scrolling marquee characters. Not implemented.
			skipbytes = 2;
			break;

		case 22: // Enable scrolling marquee. Not implemented.
			skipbytes = 3;
			break;

		case 23: // Wrap ON. Not implemented. See lcd.h to set wrap mode.
			break;

		case 24: // Wrap OFF. Not implemented. See lcd.h to set wrap mode.
			break;

		case 25: // Set custom character bitmap. Not implemented.
			skipbytes = 8;
			break;

		case 26: // Reboot
			lcdstart(); //Almost a reboot.
			
		case 27: // Escape sequence prefix. Not implemented.
			skipbytes = 2;
			break;

		case 28: // Large block number (3x4 or 4x4). Not implemented.
			skipbytes = 3;
			break;

		case 29: // Not used
			break;

		case 30: // Send data directly to the LCD controller. Extremely not implemented.
			skipbytes = 2;
			break;

		case 31: // Show information screen
			lcd_clrscr();
			lcd_puts_P("Info screen");
			break;

		// Translate from LCD Smartie pseudo-Crystal Fontz muppet-ASCII to HD44780-muppet-ASCII
		case 196: lcd_putc (1);    break;  //Upper case 
		case 214: lcd_putc (0xEF); break;	//Upper case 
		case 228: lcd_putc (0xE1); break; //Lower case 
		case 246: lcd_putc (3);    break; //Lower case 
		case 197: lcd_putc (0);    break; //Upper-case 
		case 229: lcd_putc (2);    break; //Lower case 
		case 251:  lcd_putc (4);    break; // backslash (\)


		case 127: // ASCII DEL. Ignore.
			break;

		default: 		// We have tried every non-printable character.
		lcd_putc(c);	// So the character _must_ be printable. So print it.
			break;
		}
	}
}
